//Name: Justin Fiedler
//Project: HW1
//Class: cs 352

import java.util.*;

public class Graph
{	private int nodeCount;
	private int arcCount;
	private int nodeSpace;
	private int arcSpace;

	private int[][] nodeArray;
	private final int POINT  = 0;
	private final int RPOINT = 1;
	private final int LABEL  = 2;
	private final int XCOORD = 3;
	private final int YCOORD = 4;
	private final int PARENT = 5;
	private final int CHILD  = 6;
	private final int PCOST  = 7;

	private int[][] arcArray;
	private final int TAIL  = 0;
	private final int HEAD  = 1;
	private final int COST  = 2;
	private final int TRACE = 3;

	public Graph()
	{	nodeCount = 0;
		nodeSpace = 10;
		arcCount  = 0;
		arcSpace  = 10;
		nodeArray = new int[nodeSpace][8];
		arcArray  = new int[arcSpace][4];
	}
        
        /* clearPathesAndCosts
	 * resets all nodes found parents/child/and path costs to -1
         * indicating that none exist
         */
	public void clearPathsAndCosts()
	{	for(int i = 0; i < nodeCount; i++)
		{	nodeArray[i][PARENT] = -1;
			nodeArray[i][CHILD] = -1;
			nodeArray[i][PCOST]  = -1;
		}
	}

        /* addArc
	 * adds an arc from node tail to node head of cost 'cost'
	 * only if both said nodes exist
	 */
	public void addArc(String tail, String head, int cost)
	{	tail = "" + tail.charAt(0);
		head = "" + head.charAt(0);
		if(nodeIsInGraph(tail) || nodeIsInGraph(head))
		{   addArc(findIntLabel(tail), findIntLabel(head), cost);
		}
	}
        
        /* addNode
	 * adds node of label 'label' and x, y coordinates 'x','y'
	 * to the graph
	 */
	public void addNode(String label, int x, int y)
	{	if(nodeArrayIsFull()) expandNodeArray();
		nodeArray[nodeCount][LABEL]  = (int)(label.charAt(0));
		nodeArray[nodeCount][XCOORD] = x;
		nodeArray[nodeCount][YCOORD] = y;
		nodeArray[nodeCount][PARENT] = -1;
		nodeArray[nodeCount][CHILD]  = -1;
		nodeArray[nodeCount][PCOST]  = -1;
		nodeCount++;
	}
        
        /* getArcCost
	 * returns the cost of the arc from tail to head
	 * if tail or head are not in the graph then the cost
	 * is returned as -1 indicating no arc
	 */
	public int getArcCost(String tail, String head)
	{	if(!nodeIsInGraph(tail) || !nodeIsInGraph(head))
	        {	return -1;
		}
		return getArcCost(findIntLabel(tail), findIntLabel(head));
	}

        /* getArcCost
	 * returns the cost of arc from tail to head
	 * if no such arc exists the cost is retrurned as -1
	 */
	private int getArcCost(int tail, int head)
	{	int x = nodeArray[tail][POINT];
		for(; x < arcCount && arcArray[x][TAIL] == tail && arcArray[x][HEAD] < head; x++);
		if(x > arcCount || arcArray[x][TAIL] != tail)
		{	return -1;
		}
		return arcArray[x][COST];
	}

        /* getEuclideanCost
	 * returns the least Euclidean distance (determined by
	 * tail/head x/y coordinates) between tail and head
	 */
	public double getEuclideanCost(String tail, String head)
	{	if(!hasCommas(head)) head = addCommas(head);
		if(head.length() == 1)
		{	if(!nodeIsInGraph(tail) || !nodeIsInGraph(head))
			{	return -1;
			}
			return getEuclideanCost(findIntLabel(tail), findIntLabel(head));
		}
		else
		{	double least = -1;
			StringTokenizer st = new StringTokenizer(head, ",");
			if(st.hasMoreTokens())
			{	least = getEuclideanCost(tail, st.nextToken());
			}
			while(st.hasMoreTokens())
			{	double d = getEuclideanCost(tail, st.nextToken());
				if(d < least) least = d;
			}
			return least;
		}
	}

        /* getEuclideanCost
	 * returns the least Euclidean distance (determined by
	 * tail/head x/y coordinates) between tail and head
	 */
	private double getEuclideanCost(int tail, int head)
	{	int xdiff = nodeArray[tail][XCOORD] - nodeArray[head][XCOORD];
		int ydiff = nodeArray[tail][YCOORD] - nodeArray[head][YCOORD];
		return Math.sqrt(xdiff * xdiff + ydiff * ydiff);
	}

        /* getPAthCost
	 * returns the path cost of node of label 'label'
	 * if node is not in graph cost is returned as -1
	 */
	public int getPathCost(String label)
	{	if(!nodeIsInGraph(label))
		{	return -1;
		}
		return nodeArray[findIntLabel(label)][PCOST];
	}

        /* setPathCost
	 * sets the path cost of node 'label' to pathCost
         * if node is the graph
	 */
	public void setPathCost(String label, int pathCost)
	{	if(nodeIsInGraph(label))
		{	nodeArray[findIntLabel(label)][PCOST] = pathCost;
		}
	}

        /* getChildren
	 * returns all direct children of node 'label'
	 */
	public String getChildren(String label)
	{	String children = "";
		if(nodeIsInGraph(label))
		{	int[] iA = getChildren(findIntLabel(label));
			for(int i = 0; i < iA.length; i++)
			{	children += (char)nodeArray[iA[i]][LABEL];
			}
			children = commaAlphaSort(children);
		}
		//System.out.println(label + "children: " + children);
		return children;
	}

        /* getChildren
	 * returns all direct children of node 'label'
	 */
	private int[] getChildren(int label)
	{	int[] temp = new int[10];
		int i = 0;
		int x = nodeArray[label][POINT];
		for(; x < arcCount && arcArray[x][TAIL] == label; x++, i++)
		{	if(temp.length == i) resizeArray(temp, 2*temp.length);
			temp[i] = arcArray[x][HEAD];
		}
		temp = resizeArray(temp, i);
		return temp;
	}

        /* getChildren
	 * returns all incoming nodes directly preceeding node 'label'
	 */
	public String getParents(String label)
	{	String parents = "";
		if(nodeIsInGraph(label))
		{	int[] iA = getParents(findIntLabel(label));
			for(int i = 0; i < iA.length; i++)
			{	parents += (char)nodeArray[iA[i]][LABEL];
			}
			parents = commaAlphaSort(parents);
		}
		return parents;
	}

        /* getChildren
	 * returns all incoming nodes directly preceeding node 'label'
	 */
	private int[] getParents(int label)
	{	int[] temp = new int[10];
		int i = 0;
		int x = nodeArray[label][RPOINT];
		for(; x < arcCount && arcArray[arcArray[x][TRACE]][HEAD] == label; x++, i++)
		{	if(temp.length == i) resizeArray(temp, 2*temp.length);
			temp[i] = arcArray[arcArray[x][TRACE]][TAIL];
		}
		temp = resizeArray(temp, i);
		return temp;
	}

        /* getParent
	 * returns the node set as the predacessor of node 'label'
	 * in the current path
	 */
	public String getParent(String label)
	{	String s = "";
		if(!nodeIsInGraph(label))
		{	return s;
		}
		return s += (char)(nodeArray[nodeArray[findIntLabel(label)][PARENT]][LABEL]);
	}

        /* getChild
	 * returns the node set as the next node following node 'label'
	 * in the current path
	 */
	public String getChild(String label)
	{	String s = "";
		if(!nodeIsInGraph(label))
		{	return s;
		}
		return s += (char)(nodeArray[nodeArray[findIntLabel(label)][CHILD]][LABEL]);
	}

        /* hasParent
	 * returns true if node 'label' has a predacessor in the current path
         * and false otherwise
	 */
	public boolean hasParent(String label)
	{	return nodeArray[findIntLabel(label)][PARENT] != -1;
	}

        /* hasChild
	 * returns true if node 'label' has a set decendant in the current path
         * and false otherwise
	 */
	public boolean hasChild(String label)
	{	return nodeArray[findIntLabel(label)][CHILD] != -1;
	}

        /* setPArent
	 * sets the predacessor of node 'label' in the path to 'parent'
	 */
	public void setParent(String label, String parent)
	{	nodeArray[findIntLabel(label)][PARENT] = findIntLabel(parent);
	}

        /* setChild
	 * sets the node following node 'label' in the path to 'child'
	 */
	public void setChild(String label, String child)
	{	nodeArray[findIntLabel(label)][CHILD] = findIntLabel(child);
	}

        /* getBiPath
         * returns a string representing the path from start to goal
	 * through mid by following the back path from mid to start
	 * and the forward path from mid to goal. If no path has been
	 * set the path returned is ""
	 */
	public String getBiPath(String start, String mid, String goal)
	{	int i = 0;
		String path = "";
		if(findIntLabel(mid) != -1)
		{	path += getBackPath(start, mid);
			path = path.substring(0, path.length() - 1);
			path += getForwardPath(mid, goal);
		}
		return path;
	}

        /* getForwardPath
	 * returns the path from start to goal by following the 
         * 'child' of each node.  If no path has been set "" is returned
	 */
	public String getForwardPath(String start, String goal)
	{	String path = "";
		if(goal.length() == 1)
		{	if(nodeIsInGraph(start))
			{	String node = start;
				while(!node.equals(goal) && hasChild(node))
				{	if(node.equals(start))
					{	path += node;
					}
					else
					{	path += "->" + node;
					}
					node = getChild(node);
				}
				path += "->" + node;
				if(!node.equals(goal))
				{	path = "";
				}
			}
		}
		else
		{	String token;
			StringTokenizer st = new StringTokenizer(goal, ",");
			while(st.hasMoreTokens())
			{	token = st.nextToken();
				if(getForwardPath(start, token) != "")
				{	path = getForwardPath(start, token);
				}
			}
		}
		return path;
	}

        /* getBackPath
	 * returns the path from start to goal by tracing the
	 * parents of the nodes starting at goal until node start
	 * has been reached. If no path has been set "" is returned.
	 */
	public String getBackPath(String start, String goal)
	{	String path = "";
		if(nodeIsInGraph(goal))
		{	String node = goal;
			while(!node.equals(start) && hasParent(node))
			{	path = "->" + node + path;
				node = getParent(node);
			}
			path = node + path;
			if(!node.equals(start))
			{	System.out.println("No such back path");
				System.exit(0);
			}
		}
		return path;
	}

        /* getPathTo
	 * returns the longest traceable path to node 'label'
	 * by following the parent of each node until no
	 * parent exists
	 */
	public String getPathTo(String label)
	{	String path = "";
		if(nodeIsInGraph(label))
		{	path = label;
			if(hasParent(label))
			{	path = getPathTo(getParent(label)) + "->" + path;
			}
		}
		return path;
	}

        /* addArc
	 * adds and arc from tail to head of cost 'cost'
         * and reorders the structure.
	 */
	private void addArc(int tail, int head, int cost)
	{	if(arcArrayIsFull()) expandArcArray();

		int x = nodeArray[tail][POINT];
		for(; x < arcCount && arcArray[x][TAIL] == tail && arcArray[x][HEAD] < head; x++);

		for(int i = arcCount - 1; i >= x; i--)
		{		for(int j = 0; j < 4; j++)
			{	arcArray[i + 1][j] = arcArray[i][j];
			}
		}

		arcArray[x][TAIL] = tail;
		arcArray[x][HEAD] = head;
		arcArray[x][COST] = cost;
		arcCount++;

		setRPointTrace();
	}

        /* findIntLabel
	 * returns the array position of the node with label 'label'
	 * or -1 if no such node exists
	 */
	private int findIntLabel(String label)
	{	if(label == null)
		{	return -1;
		}
		return findIntLabel(label.charAt(0));
	}
    
        /* findIntLabel
	 * returns the array position of the node with label 'label'
	 * or -1 if no such node exists
	 */
	private int findIntLabel(char label)
	{	int intLabel = (int)(label);
		for(int i = 0; i < nodeCount; i++)
		{	if(nodeArray[i][LABEL] == intLabel)
				return i;
		}
		return -1;
	}

        /* returns true if node of label 'node' is in the graph
	 * or false otherwise
	 */
	public boolean nodeIsInGraph(String node)
        {        return findIntLabel(node) != -1;
	}

        /* setRPointTrace
	 * reorganizes the arrays into star representation
	 */
	private void setRPointTrace()
	{	int tPos = 0;
		int pPos = 0;
		for(int k = 0; k < nodeCount; k++)
		{	nodeArray[k][POINT]  = pPos;
			nodeArray[k][RPOINT] = tPos;
			for(int i = 0; i < arcCount; i++)
			{	if(arcArray[i][TAIL] == k)
				{	//arcArray[pPos][POINT] = i;
					pPos++;
				}
				if(arcArray[i][HEAD] == k)
				{	arcArray[tPos][TRACE] = i;
					tPos++;
				}
			}
		}
	}

        /* arcArrayIsFull
	 * returns true if arcArray is full
	 */
	private boolean arcArrayIsFull()
	{	return arcCount == arcSpace;
	}

        /* expandArcArray
	 * doubles the capacity of arcArray
	 */
	private void expandArcArray()
	{	arcSpace += 10;
		arcArray = expand2DArray(arcArray, arcSpace, arcArray[0].length);
	}

        /* nodeArrayIsFull
	 * returns true if arcArray is full
	 */
	private boolean nodeArrayIsFull()
	{	return nodeCount == nodeSpace;
	}

        /* expandNodeArray
	 * doubles the capacity of nodeArray
	 */
	private void expandNodeArray()
	{	nodeSpace += 10;
		nodeArray = expand2DArray(nodeArray, nodeSpace, nodeArray[0].length);
	}

        /* resizeArray
	 * returns an int array of size 'newSize' containing the 
	 * contents of 'arr'
	 */
	private int[] resizeArray(int[] arr, int newSize)
	{	int[] temp = new int[newSize];
		for(int i = 0; (i < arr.length && i < newSize); i++)
		{	temp[i] = arr[i];
		}
		return temp;
	}

        /* expand2DArray
	 * expands a 2 dimensional int array to size [x][y]
	 */
	private int[][] expand2DArray(int[][] arr, int x, int y)
	{	int[][] temp = new int[x][y];
		for(int i = 0; i < arr.length; i++)
		{	for(int j = 0; j < arr[0].length; j++)
			{	temp[i][j] = arr[i][j];
			}
		}
		return temp;
	}

        /* commaAlphaSort
	 * alphabetically sorts nodeString and then adds commas
	 * between nodes for easy tokenizing
	 */
	public static String commaAlphaSort(String nodeString)
	{	return addCommas(alphaSort(nodeString));
	}

        /* alphaSort
	 * alphabetically sorts nodeString ignoring commas
	 */
	public static String alphaSort(String nodeString)
	{	boolean commas = hasCommas(nodeString);
		if(commas) nodeString = removeCommas(nodeString);
		char[] ch = nodeString.toCharArray();
		for(int i = 0; i < nodeString.length() - 1; i++)
    	{	for(int j = 0; j < nodeString.length() - i - 1; j++)
    		{	if(ch[j + 1] < ch[j])
    			{	char temp = ch[j];
    				ch[j] = ch[j + 1];
    				ch[j + 1] = temp;
    			}
    		}
    	}
    	String temp = "";
    	for(int i = 0; i < nodeString.length(); i++)
    	{	temp += ch[i];
    	}
    	if(commas) temp = addCommas(temp);
    	return temp;
	}

        /* hasCommas
	 * returns true if nodeString contains commas or false otherwise
	 */
	public static boolean hasCommas(String nodeString)
	{	return nodeString.indexOf(",") != -1;
	}

        /* addCommas
	 * seperates each "node" in node string by a comma
	 */
	public static String addCommas(String nodeString)
	{	boolean commas = hasCommas(nodeString);
		if(commas) nodeString = removeCommas(nodeString);
		char[] ch = nodeString.toCharArray();
		String temp = "";
    	for(int i = 0; i < nodeString.length(); i++)
    	{	temp += ch[i];
    		if(i < nodeString.length() - 1) temp += ",";
    	}
    	return temp;
	}

        /* removeCommas
	 * removes all commas(',') from nodeString. why? because we hate them.
	 */
	public static String removeCommas(String nodeString)
	{	String temp = "";
    	for(int i = 0; i < nodeString.length(); i++)
    	{	char c = nodeString.charAt(i);
    		if(c != ',') temp += c;
    	}
    	return temp;
	}

        /* toString
	 * returns a String representation of the graph
	 */
	public String toString()
	{	String s = "Node Array\n";
		for(int i = 0; i < nodeCount; i++)
		{	for(int j = 0; j < nodeArray[0].length; j++)
			{	if(j == LABEL)
				{	s += (char)(nodeArray[i][j]) + " ";
				}
				else
				{	s += nodeArray[i][j] + " ";
				}
			}
			s += "\n";
		}

		s += "\nArc Array\n";
		for(int i = 0; i < arcCount; i++)
		{	for(int j = 0; j < arcArray[0].length; j++)
			{	if(j == TAIL || j== HEAD)
				{	s += (char)(nodeArray[arcArray[i][j]][LABEL]) + " ";
				}
				else
				{	s += arcArray[i][j] + " ";
				}
			}
			s += "\n";
		}
		return s;
	}
}
