//Name: Justin Fiedler
//Project: HW1
//Class: cs352

import java.io.*;
import java.util.*;

public class HW1
{	Graph g;
        boolean debug;

	public static void main(String args[])
        {	boolean d = false;
	        String graphFile = args[0];
	        String searchFile = args[1];
		//if debug is on
		if(args.length == 3  && args[2].equals("-d"))
		{    d = true;
		}

		new HW1(graphFile, searchFile, d);
	}

	public HW1(String graphFile, String searchFile, boolean debug)
	{	this.debug = debug;
		readGraphFile(graphFile);
		readSearchFile(searchFile);
	}
        
        //function that accepts the name of a file to be read as a graph
        //read in file using MyFileReader
	public void readGraphFile(String graphFile)
	{	g =  new Graph();

		String label, tail, head;
		int x, y, cost;

		try
		{	MyFileReader fr = new MyFileReader(graphFile, ",");
			String line = fr.readLine();
			while(line != null)
			{	label = fr.nextToken();
				x = Integer.parseInt(fr.nextToken());
				y = Integer.parseInt(fr.nextToken());
				g.addNode(label, x, y);
				line = fr.readLine();
			}

			line = fr.readLine();
			while(line != null)
			{	tail = fr.nextToken();
				head = fr.nextToken();
				cost = Integer.parseInt(fr.nextToken());
				g.addArc(tail, head, cost);
				line = fr.readLine();
			}
		}
		catch(FileNotFoundException e)
		{	System.out.println(graphFile + " not found");
			System.exit(0);
		}
		catch(IOException e)
		{	System.out.println("Error reading from " + graphFile);
			System.exit(0);
		}
	}

        //function that accepts the name of a file to be read as a list of searches
        //read in file using MyFileReader
	public void readSearchFile(String searchFile)
	{	String searchType, start, goal;
		try
		{	MyFileReader fr = new MyFileReader(searchFile, "\t,");
			String line = fr.readLine();
			while(line != null)
			{	searchType = start = goal = "";
				searchType = fr.nextToken();
				start = fr.nextToken();
				while(fr.hasMoreTokens())
				{	goal += fr.nextToken();
				}

				search(searchType, start, goal);

				line = fr.readLine();
			}
		}
		catch(FileNotFoundException e)
		{	System.out.println(searchFile + " not found");
			System.exit(0);
		}
		catch(IOException e)
		{	System.out.println("Error reading from " + searchFile);
			System.exit(0);
		}
	}
        
        /* search
         * a function to perform the search specified by 'searchType' on graph 'g'
         * from node 'start' to node(s) 'goal'
	 */
	public void search(String searchType, String start, String goal)
	{	g.clearPathsAndCosts();
		System.out.println(searchType + " from " + start + " to " + Graph.addCommas(goal));
		System.out.println("--------------------------------------");

		if(searchType.equals("BFS"))
		{	System.out.println("Solution: " + BFS(start, goal) + "\n\n");
		}
		else if(searchType.equals("UCS"))
		{	System.out.println("Solution: " + UCS(start, goal) + "\n\n");
		}
		else if(searchType.equals("GS"))
		{	System.out.println("Solution: " + GS(start, goal) + "\n\n");
		}
		else if(searchType.equals("AS"))
		{	System.out.println("Solution: " + AS(start, goal) + "\n\n");
		}
		else if(searchType.equals("DFS"))
		{	System.out.println("Solution: " + DFS(start, goal) + "\n\n");
		}
		else //(searchType.equals("BDS"))
		{	System.out.println("Solution: " + BDS(start, goal) + "\n\n");
		}
		System.out.println();
	}

	/* Breadth First Search
         * performs a BFS search on graph 'g'
         * from node 'start' to node(s) 'goal'
	 */
	public String BFS(String start, String goal)
	{	String closedList = "";
		Queue OL = new Queue();

		OL.enqueue(start);
		g.setPathCost(start, 0);

		while(!OL.isEmpty())
		{	String n = (String)(OL.dequeue());
			//debug print
			if(debug) System.out.println("TEST: " + n);
			if(!isIn(closedList, n))
			{	closedList += n;
				if(isIn(goal,n))
				{	return g.getBackPath(start, n);
				}
				else
				{	String children = g.getChildren(n);
					StringTokenizer st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList, x))
						{	OL.enqueue(x);
							if(!g.hasParent(x))
							{	g.setParent(x, n);
								g.setPathCost(x, g.getPathCost(n) + 1);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST: " + OL);
				System.out.println("CLOSED LIST: " + Graph.commaAlphaSort(closedList));
				System.out.println();
			}
		}
		System.out.println("TEST: null");
		return new String("There is no solution path.");
	}

	/* Uniform Cost Search
         * performs a UCS on graph 'g'
         * from node 'start' to node(s) 'goal'
	 */
	public String UCS(String start, String goal)
	{	String closedList = "";
		PriorityQueue OL = new PriorityQueue();

		OL.enqueue(start, 0);
		g.setPathCost(start, 0);

		String n = (String)OL.dequeue();;
		while(n != null)
		{	//debug print
			if(debug) System.out.println("TEST: " + n);
			if(!isIn(closedList, n))
			{	closedList += n;
				if(isIn(goal, n))
				{	return g.getBackPath(start, n);
				}
				else
				{	String children = g.getChildren(n);
					StringTokenizer st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList, x))
						{	int gx = g.getPathCost(n) + g.getArcCost(n, x);
							OL.enqueue(x, gx);
							if(gx < g.getPathCost(x) || g.getPathCost(x) == -1)
							{	g.setParent(x, n);
								g.setPathCost(x, gx);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST: " + OL);
				System.out.println("CLOSED LIST: " + Graph.commaAlphaSort(closedList));
				System.out.println();
			}
			n = (String)OL.dequeue();
		}
		System.out.println("TEST: null");
		return new String("There is no solution path.");
	}

        /*Greedy Search
         * performs a GS on graph 'g'
	 * from node 'start' to node(s) 'goal'
	 */
	public String GS(String start, String goal)
	{	String closedList = "";
		PriorityQueue OL = new PriorityQueue();

		OL.enqueue(start, 0);
		g.setPathCost(start, 0);
		
		String n = (String)OL.dequeue();;
		while(n != null)
		{	//debug print
			if(debug) System.out.println("TEST: " + n);
			if(!isIn(closedList, n))
			{	closedList += n;
				if(isIn(goal, n))
				{	return g.getBackPath(start, n);
				}
				else
				{	String children = g.getChildren(n);
					StringTokenizer st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList, x))
						{	double hx = g.getEuclideanCost(x, goal);
							OL.enqueue(x, hx);
							if(!g.hasParent(x))
							{	g.setParent(x, n);
								g.setPathCost(x, g.getPathCost(n) + 1);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST: " + OL);
				System.out.println("CLOSED LIST: " + Graph.commaAlphaSort(closedList));
				System.out.println();
			}
                        n = (String)OL.dequeue();
		}
		System.out.println("TEST: null");
		return new String("There is no solution path.");
	}

        /* A Search
         * performs an AS on graph 'g'
	 * from node 'start' to node(s) 'goal'
	 */
	public String AS(String start, String goal)
	{	String closedList = "";
		PriorityQueue OL = new PriorityQueue();

		OL.enqueue(start, 0);
		g.setPathCost(start, 0);

		while(OL.getHighestPriority() != -1)
		{	String n = (String)OL.dequeue();
			//debug print
			if(debug) System.out.println("TEST: " + n);
			if(!isIn(closedList, n))
			{	closedList += n;
				if(isIn(goal, n))
				{	return g.getBackPath(start, n);
				}
				else
				{	String children = g.getChildren(n);
					StringTokenizer st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList, x))
						{	int gx = g.getPathCost(n) + g.getArcCost(n, x);
							double hx = g.getEuclideanCost(x, goal);
							OL.enqueue(x, gx + hx);
							if(gx < g.getPathCost(x) || g.getPathCost(x) == -1)
							{	g.setParent(x, n);
								g.setPathCost(x, gx);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST: " + OL);
				System.out.println("CLOSED LIST: " + Graph.commaAlphaSort(closedList));
				System.out.println();
			}
		}
		System.out.println("TEST: null");
		return new String("There is no solution path.");
	}

        /* Depth First Search
         * performs an DFS on graph 'g'
	 * from node 'start' to node(s) 'goal'
	 */
	public String DFS(String start, String goal)
	{	String closedList = "";
		Stack OL = new Stack();

		OL.push(start);
		g.setPathCost(start, 0);

		while(!OL.isEmpty())
		{	String n = (String)OL.pop();
			//debug print
			if(debug) System.out.println("TEST: " + n);
			if(!isIn(closedList, n))
			{	closedList += n;
				if(isIn(goal, n))
				{	return g.getBackPath(start, n);
				}
				else
				{	String children = g.getChildren(n);
					children = reverseNodeString(children);
					StringTokenizer st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList, x))
						{	OL.push(x);
							g.setParent(x, n);
							g.setPathCost(x, g.getPathCost(n) + 1);
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST: " + OL);
				System.out.println("CLOSED LIST: " + Graph.commaAlphaSort(closedList));
				System.out.println();
			}
		}
		System.out.println("TEST: null");
		return new String("There is no solution path.");
	}

        /* Bidirectional Breadth First Search
         * performs an BDS on graph 'g'
	 * from node 'start' to node(s) 'goal'
	 */
	public String BDS(String start, String goal)
	{	String[] closedList = new String[2];
		String[] goalSet    = new String[2];
		Queue[] OL = new Queue[2];

		for(int i = 0; i < 2; i++)
		{	closedList[i] = "";
			goalSet[i] = "";
			OL[i] = new Queue();
		}
		
		OL[0].enqueue(start);	//Add start node to OL1
		
		goal = Graph.commaAlphaSort(goal);

		StringTokenizer st = new StringTokenizer(goal, ",");
		while(st.hasMoreTokens())
		{	OL[1].enqueue(st.nextToken());  //Add goal nodes to OL2
		}

		goalSet[0] += goal;	//Add goal nodes to GN1
		goalSet[1] += start;	//Add start node to GN2

		String n1, n2;
		if(!OL[0].isEmpty() || !OL[1].isEmpty())
		{
                do
		{	n1 = (String)OL[0].dequeue();
			//debug print
			if(debug) System.out.println("TEST: " + n1);
			if(!isIn(closedList[0], n1))	//n is not on CL1
			{	closedList[0] += n1;		//add n to CL1
				goalSet[1] += n1;		//add n to GN2
				if(isIn(goalSet[0], n1))	//n is in GN1
				{	return g.getBiPath(start, n1, goal);
				}
				else
				{	String children = g.getChildren(n1);
					st = new StringTokenizer(children, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList[0], x))
						{	OL[0].enqueue(x);
							if(!g.hasParent(x))
							{	g.setParent(x, n1);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST 1: " + OL[0]);
				System.out.println("CLOSED LIST 1: " + Graph.commaAlphaSort(closedList[0]));
				System.out.println();
			}

			n2 = (String)OL[1].dequeue();
			//debug print
			if(debug) System.out.println("TEST: " + n2);
			if(!isIn(closedList[1], n2))	//n is not on CL1
			{	closedList[1] += n2;		//add n to CL1
				goalSet[0] += n2;		//add n to GN2
				if(isIn(goalSet[1], n2))	//n is in GN1
				{	return g.getBiPath(start, n2, goal);
				}
				else
				{	String parents = g.getParents(n2);
					st = new StringTokenizer(parents, ",");
					while(st.hasMoreTokens())
					{	String x = st.nextToken();
						if(!isIn(closedList[1], x))
						{	OL[1].enqueue(x);
							if(!g.hasChild(x))
							{	g.setChild(x, n2);
							}
						}
					}
				}
			}
			//debug print
			if(debug)
			{	System.out.println("OPEN LIST 2: " + OL[1]);
				System.out.println("CLOSED LIST 2: " + Graph.commaAlphaSort(closedList[1]));
				System.out.println();
			}
		} while (n1 != null || n2 != null);
		}
		return new String("There is no solution path.");
	}

        /* isIn
         * returns true is there is any instance of 'subString' in 's'
         * note: null will always return true as any string contains null/nothing
         */ 
	private boolean isIn(String s, String subString)
	{	if(subString == null)
		{	return true;
		}
		return s.indexOf(subString) != -1;
	}
        
        /* reverseNodeString
         * returns the reverses the node order of node string 's', if there are commas they
         * properly placed for tokenizing, if there are no commas then they're wont be commas
         */ 
	private String reverseNodeString(String s)
	{	boolean commas = false;
		if(Graph.hasCommas(s))
		{	commas = true;
			s = Graph.removeCommas(s);
		}
		char[] ch = s.toCharArray();
		for(int i  = 0; i < s.length()/2; i++)
		{	char temp = ch[i];
			ch[i] = ch[s.length() - i - 1];
			ch[s.length() - i - 1] = temp;
		}
		if(commas)	s = Graph.addCommas(new String(ch));
		return s;
	}
}
